import math
from typing import Dict, Any, Tuple, Set, Callable, Optional, Union
from reactivex import operators as ops, Observable, from_iterable
from reactivex import just, empty
from pybeamline.boevent import BOEvent
from pybeamline.models.aer import AER
from pybeamline.utils.cardinality import infer_cardinality, Cardinality


def activity_entity_relations_miner_lossy_counting(model_update_frequency=10, max_approx_error: float = 0.01, control_flow: Optional[Set[str]] = None) -> Callable[
    [Observable[BOEvent]], Observable[AER]]:
    """
    Entity Relationship Miner using a lossy counting approach.
    :param control_flow:
    Optional dictionary mapping object types (str) to miner factory callables (i.e., `Callable[[], StreamMiner]`).
    :param model_update_frequency:
    Frequency of model updates
    :param max_approx_error:
     Maximum approximation error for the lossy counting on objects
    :return: Activity-Entity Relationship (AER)
    """
    obj_rel = ActivityEntityRelationMinerLossyCounting(max_approx_error=max_approx_error, control_flow=control_flow)

    def miner(event: Union[BOEvent,dict]) -> Observable[AER]:
        obj_rel.ingest_event(event)
        if obj_rel.observed_events() % model_update_frequency == 0:
             return just(obj_rel.get_model())
        else:
            return empty()

    return ops.flat_map(miner)

class ActivityEntityRelationMinerLossyCounting:
    def __init__(self, max_approx_error: float = 0.001, control_flow: Optional[Set[str]] = None):
        self.__control_flow = control_flow
        self.__D_C: Dict[str, Dict[Tuple[str, str], Dict[Cardinality, Tuple[int, int]]]] = {}
        self.__D_O: Dict[str, Set[str]] = {}
        self.__D_N: Dict[str, int] = {}
        self.__observed_events = 1
        self.__bucket_width = int(1 / max_approx_error)

    def ingest_event(self, event: BOEvent):
        activity = event.get_event_name()
        omap = event.get_omap()

        self.__D_N[activity] = self.__D_N.get(activity, 0) + 1

        current_bucket = math.ceil(self.__D_N[activity] / self.__bucket_width)
        obj_types = [t for t in omap if not self.__control_flow or t in self.__control_flow]

        self.__D_O.setdefault(activity, set()).update(obj_types)

        if len(obj_types) >= 2:
            rel_map = self.__D_C.setdefault(activity, {})
            for i in range(len(obj_types)):
                for j in range(i + 1, len(obj_types)):
                    type1, type2 = sorted([obj_types[i], obj_types[j]])
                    key = (type1, type2)
                    count1, count2 = len(omap[type1]), len(omap[type2])
                    card = infer_cardinality(count1, count2)

                    card_map = rel_map.setdefault(key, {})
                    if card in card_map:
                        freq, delta = card_map[card]
                        card_map[card] = (freq + 1, delta)
                    else:
                        card_map[card] = (1, current_bucket)

        if self.__D_N[activity] % self.__bucket_width == 0:
            self._cleanup(current_bucket, activity)

        self.__observed_events+= 1

    def _cleanup(self, current_bucket: int, activity: str):
        if activity not in self.__D_C:
            return
        rel_map = self.__D_C[activity]

        for key, card_map in rel_map.items():
            to_remove_cards = [c for c, (freq, delta) in card_map.items() if freq + delta <= current_bucket]
            for c in to_remove_cards:
                del card_map[c]

    def get_model(self) -> AER:
        diagram = AER()
        for activity, obj_types in self.__D_O.items():
            diagram.add_object_types(activity, obj_types)

        for activity, rel_map in self.__D_C.items():
            for (obj1, obj2), card_map in rel_map.items():
                most_common_card = max(card_map.items(), key=lambda kv: kv[1][0])[0]
                diagram.add_relation(activity, obj1, obj2, most_common_card)

        return diagram

    def observed_events(self) -> int:
        return self.__observed_events





import random
from datetime import datetime
from typing import List, Tuple, Optional, Dict
from reactivex import Observable, abc
from reactivex.disposable import CompositeDisposable
from pybeamline.boevent import BOEvent


def generate_shuffled_traces(flows: List[Tuple[List[dict], int]], shuffle: bool = True) -> List[dict]:
    all_traces = []
    trace_id = 0

    for flow_template, repetitions in flows:
        for _ in range(repetitions):
            trace = []
            suffix = f"{trace_id}"
            for event in flow_template:
                updated_event = {
                    "activity": event["activity"],
                    "objects": {
                        obj_type: [f"{oid}_{suffix}" for oid in obj_ids]
                        for obj_type, obj_ids in event["objects"].items()
                    }
                }
                trace.append(updated_event)
            all_traces.append(trace)
            trace_id += 1

    if shuffle:
        random.shuffle(all_traces)

    return [event for trace in all_traces for event in trace]


def dict_test_ocel_source(flows: List[Tuple[List[dict], int]], shuffle: bool = False, scheduler: Optional[abc.SchedulerBase] = None) -> Observable[BOEvent]:
    """
    :param flows: A list of tuples, where each tuple is of the form (flow_template, repetitions).
                  - flow_template is a list of event dictionaries with keys "activity" and "objects".
                  - repetitions is the number of traces to generate from that template.
                  Example:
                  [
                      (
                          [
                              {"activity": "Register Customer", "objects": {"Customer": ["c1"]}},
                              {"activity": "Create Order", "objects": {"Customer": ["c1"], "Order": ["o1"]}}
                          ],
                          20
                      )
                  ]
                  This means: generate 20 traces with those two events.
    :param shuffle: Whether to shuffle the events from different traces in the final output.
    :param scheduler: (Optional) A ReactiveX scheduler to control event emission timing.
    :return: An Observable stream of BOEvent objects, one per event.
    """
    def subscribe(observer: abc.ObserverBase[BOEvent], scheduler_: Optional[abc.SchedulerBase] = None) -> abc.DisposableBase:
        all_events = generate_shuffled_traces(flows, shuffle=shuffle)
        for idx, event in enumerate(all_events):

            omap: Dict[str,set[str]] = {
                obj_type: set(ids)
                for obj_type, ids in event["objects"].items()
            }

            bo_event = BOEvent(
                event_id=f"e{idx}",
                activity_name=event["activity"],
                timestamp=datetime.now(),
                omap=omap,
                vmap=None,
            )
            observer.on_next(bo_event)

        observer.on_completed()
        return CompositeDisposable()

    return Observable(subscribe)





from typing import List, Callable
from reactivex import Observable, empty, just
from reactivex import operators as ops

from pybeamline.abstractevent import AbstractEvent
from pybeamline.bevent import BEvent
from pandas import DataFrame

def list_to_log(events: List[AbstractEvent]) -> DataFrame:
    return DataFrame([e.to_dict() for e in events])


def sliding_window_to_log() -> Callable[[Observable[Observable[AbstractEvent]]], Observable[DataFrame]]:
    def o2l(obs: Observable[AbstractEvent]) -> Observable[DataFrame]:
        return obs.pipe(
            ops.to_iterable(),  # Converts Observable[AbstractEvent] → Iterable[AbstractEvent]
            ops.flat_map(lambda x: empty() if not x else just(list_to_log(x)))
        )

    return ops.flat_map(o2l)


from typing import Callable
from pm4py.algo.discovery.heuristics.variants.classic import calculate as compute_dfg
from pm4py.objects.heuristics_net.obj import HeuristicsNet
from reactivex import Observable
from reactivex import just, empty
from reactivex import operators as ops
from pybeamline.abstractevent import AbstractEvent
from pybeamline.bevent import BEvent
from pybeamline.boevent import BOEvent


def heuristics_miner_lossy_counting_budget(
        model_update_frequency=10,
        budget=100,
        dependency_threshold=0.5,
        and_threshold=0.8) -> Callable[[Observable], Observable[HeuristicsNet]]:
    hm = HeuristicsMinerLossyCountingBudget(budget=budget, dependency_threshold=dependency_threshold,
                                            and_threshold=and_threshold)


    def miner(event: AbstractEvent) -> Observable[HeuristicsNet]:
        if isinstance(event, BOEvent):
            # Verify that the event is flattened
            if len(event.get_object_ids()) != 1:
                raise ValueError("BOEvent should be flattened before supplied to miner")

            # Wrapping BOEvent into BEvent
            trace_name = event.get_object_ids()[0]
            b_event = BEvent(
                activity_name=event.get_event_name(),
                case_id=trace_name,
                event_time=event.get_event_time()
            )
            hm.ingest_event(b_event)

        elif isinstance(event, BEvent):
            hm.ingest_event(event)

        else:
            raise TypeError(f"Unsupported event type: {type(event)}")

        if hm.observed_events() % model_update_frequency == 0:
            return just(hm.get_model())
        else:
            return empty()

    return ops.flat_map(miner)




# Class originally developed by Magnus Frederiksen as part of his BSc project at DTU entitled
# "Development of Process Mining and Complex Event Processing using Python"
class HeuristicsMinerLossyCountingBudget:
    def __init__(self, budget=10, dependency_threshold=0, and_threshold=0.8):
        self.__budget = int(budget)  # max length of stored events and relations
        self.__minimum_dependency_threshold = dependency_threshold
        self.__and_threshold = and_threshold

        self.__D_C = dict()  # set of event
        self.__D_R = dict()  # set of relations
        self.__observed_events = 1
        self.__current_bucket = 0

    def ingest_event(self, event):

        # for budget lossy counting, if relation or caseID already exists, the memory will just be replaced and not
        # expanded so we only clean up when the caseID or relation doesnt exist, which results in creating a new one
        if event.get_trace_name() in self.__D_C:  # if caseID already exist
            lastEvent = self.__D_C[event.get_trace_name()]  # localy save former event
            del self.__D_C[event.get_trace_name()]  # replace caseID's former event with new event
            self.__D_C[event.get_trace_name()] = [event.get_event_name(), lastEvent[1] + 1, lastEvent[2],
                                                  event.get_event_time()]

            r_N = (lastEvent[0], event.get_event_name())  # save relation localy

            if r_N in self.__D_R:  # if relation exists in set
                lastRelation = self.__D_R[r_N]  # localy save former relation
                del self.__D_R[r_N]  # replace relation

                diff = (event.get_event_time() - lastEvent[3]) - lastRelation[
                    2]  # incase the user wants the time for each relation
                newTime = lastRelation[2] + (diff / (lastRelation[0] + 1))

                self.__D_R[r_N] = [lastRelation[0] + 1, lastRelation[1], newTime]

            else:  # the relation doesent exist, create it
                while len(self.__D_R) + len(
                        self.__D_C) >= self.__budget:  # if budget is reached when adding a new key + iten
                    self.__bucket_cleaning()  # bucket cleaning time
                self.__D_R[r_N] = (1, self.__current_bucket, event.get_event_time() - lastEvent[3])

        else:  # caseID doesnt exist, create it
            while ((len(self.__D_R) + len(
                    self.__D_C)) >= self.__budget):  # if budget is reached when adding a new key + iten
                self.__bucket_cleaning()  # bucket cleaning time
            self.__D_C[event.get_trace_name()] = (
            event.get_event_name(), 1, self.__current_bucket, event.get_event_time())

        # clean up
        self.__observed_events += 1

    def __bucket_cleaning(self):
        self.__current_bucket += 1  # increment bucket to clean all items not within the new bucket number

        D_Ctobedel = []  # the 2 lists are needed to avoid messing with the comming loops
        D_Rtobedel = []

        for caseID, (eventName, frequency, bucket, time) in self.__D_C.items():  # for all caseIDs' occured
            if frequency + bucket <= self.__current_bucket:  # if not above the bucket threshold on all events occured
                D_Ctobedel.append(caseID)

        for caseID in D_Ctobedel:  # deleted the event
            del self.__D_C[caseID]

        for relation, (frequency, bucket, time) in self.__D_R.items():  # for all relations occured
            if frequency + bucket <= self.__current_bucket:  # if not above the bucket threshold of all relations
                D_Rtobedel.append(relation)

        for relation in D_Rtobedel:  # delete the relations
            del self.__D_R[relation]

    def get_model(self):
        dfg = dict()
        for (A, B), (frequency, bucket, time) in self.__D_R.items():
            dfg[(A, B)] = frequency
        hm = HeuristicsNet(dfg)
        return compute_dfg(hm, dependency_thresh=self.__minimum_dependency_threshold,
                           and_measure_thresh=self.__and_threshold)

    def observed_events(self):  # returns map of events observed
        return self.__observed_events - 1



import math
from typing import Optional, Dict, Set, Any
from reactivex import Observable, abc
from reactivex.disposable import CompositeDisposable
from pm4py import OCEL, ocel_sort_by_additional_column, read_ocel2
from pybeamline.boevent import BOEvent

def ocel2_log_source_from_file(log_path: str, ) -> Observable[BOEvent]:
    """
    Loads an OCEL 2.0 log from a file path and returns it as an Observable of BOEvent objects.
    :param log_path: str
    :return: Observable[BOEvent]
    """
    return ocel2_log_source(read_ocel2(log_path))


def ocel2_log_source(
    log: OCEL,
    scheduler: Optional[abc.SchedulerBase] = None
) -> Observable[BOEvent]:
    """
    Converts an OCEL object into an Observable stream of BOEvent objects,
    ordered by timestamp if available.
    """
    if log.event_timestamp is not None:
        log = ocel_sort_by_additional_column(log, "ocel:timestamp")

    def subscribe(
        observer: abc.ObserverBase[BOEvent],
        scheduler_: Optional[abc.SchedulerBase] = None
    ) -> abc.DisposableBase:

        for _, event in log.get_extended_table().iterrows():

            # Build the omap using columns that start with "ocel:type:" and are not NaN
            omap = {
                col.split("ocel:type:")[1]: event[col]
                for col in event.keys()
                if col.startswith("ocel:type:") and event[col] is not None
                and not (isinstance(event[col], float) and math.isnan(event[col]))
            }

            vmap: Dict[str,Any] = {
                col: event[col]
                for col in event.keys()
                if col not in {"ocel:eid", "ocel:activity", "ocel:timestamp"} and not col.startswith("ocel:type:")
                and event[col] is not None
                and not (isinstance(event[col], float) and math.isnan(event[col]))
            }

            bo_event = BOEvent(
                event_id=event["ocel:eid"],
                activity_name=event["ocel:activity"],
                timestamp=event["ocel:timestamp"],
                omap=omap,
                vmap=vmap
            )
            observer.on_next(bo_event)

        observer.on_completed()
        return CompositeDisposable()

    return Observable(subscribe)



import reactivex
from pybeamline.sources.real.ais import _fetch_ais_source
from pybeamline.sources.real.wikimedia import _fetch_wikimedia_stream
from pybeamline.sources.real.rejseplanen import _fetch_rejseplanen_trains


def wikimedia_source():
    return reactivex.create(_fetch_wikimedia_stream)


def ais_source(host='153.44.253.27', port=5631):
    def fetch_with_params(observer, _):
        return _fetch_ais_source(observer, host, port)
    return reactivex.create(fetch_with_params)


def rejseplanen_source():
    return reactivex.create(_fetch_rejseplanen_trains)









from abc import ABC, abstractmethod

# Facilitates the work of an interface for event handling in a system.
class AbstractEvent(ABC):

    @abstractmethod
    def get_event_name(self):
        """
        Get the name of the event.
        """
        pass

    @abstractmethod
    def get_event_time(self):
        """
        Get the timestamp of the event.
        """
        pass

    @abstractmethod
    def to_dict(self):
        """
        Convert the event to a dictionary representation.
        """
        pass


from datetime import datetime
from typing import List, Optional, Dict, Set
from pybeamline.abstractevent import AbstractEvent

DEFAULT_EVENT_ID = "ocel:eid"
DEFAULT_EVENT_ACTIVITY = "ocel:activity"
DEFAULT_EVENT_TIMESTAMP = "ocel:timestamp"
OCEL_OMAP_KEY = "ocel:omap"
DEFAULT_OBJECT_ID = "ocel:oid"
DEFAULT_OBJECT_TYPE = "ocel:type"
OCEL_VMAP_KEY = "ocel:vmap"


class BOEvent(AbstractEvent):

    def __init__(self, event_id: str, activity_name: str, omap: Dict[str, Set[str]], timestamp: Optional[datetime] = None, vmap: Optional[Dict[str, str]] = None):
        """
        Represents a single event aligned with OCEL specification.
        :param event_id: Unique event identifier (ocel:eid)
        :param activity_name: Name of the activity (ocel:activity)
        :param timestamp: Timestamp of the event (ocel:timestamp)
        :param omap: List of associated objects (ocel:omap)
        :param vmap: Additional event attributes (ocel:vmap)
        """
        self.event_id = event_id
        self.activity_name = activity_name
        self.timestamp = timestamp or datetime.now()
        self.omap = omap
        self.vmap = vmap or {}


    def get_event_id(self):
        return self.event_id

    def get_event_name(self):
        return self.activity_name

    def get_event_time(self):
        return self.timestamp

    def get_object_ids(self) -> List[str]:
        return [oid for ids in self.omap.values() for oid in ids]

    def get_omap_types(self):
        return list(self.omap.keys())

    def get_omap(self):
        return self.omap

    def get_vmap(self):
        return self.vmap

    def __str__(self):
        return str(self.to_dict())

    def __repr__(self):
        return self.__str__()

    def to_dict(self):
        return {
            DEFAULT_EVENT_ID: self.event_id,
            DEFAULT_EVENT_ACTIVITY: self.activity_name,
            DEFAULT_EVENT_TIMESTAMP: self.timestamp,
            OCEL_OMAP_KEY: self.omap,
            OCEL_VMAP_KEY: self.vmap
        }